
;
; Bind Mount Extensions 
;
; this patch adds some functionality to the --bind
; type of vfs mounts.
;
; (C) 2003-2004 Herbert Ptzl <herbert@13thfloor.at>
;
; Changelog:  
;
;   0.01  - readonly bind mounts
;   0.02  - added ro truncate handling
;         - added ro (f)chown, (f)chmod handling
;   0.03  - added ro utime(s) handling
;         - added ro access and *_ioctl
;   0.04  - added noatime and nodiratime
;         - made autofs4 update_atime uncond
;   0.05  - rewrite regarding (a)time
;         - adapted to new 2.6 vfs
;
; this patch is free software; you can redistribute it and/or
; modify it under the terms of the GNU General Public License
; as published by the Free Software Foundation; either version 2
; of the License, or (at your option) any later version.
; 
; this patch is distributed in the hope that it will be useful,
; but WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
; GNU General Public License for more details.
; 

diff -NurpP --minimal linux-2.6.8.1/arch/sparc64/solaris/fs.c linux-2.6.8.1-bme0.05/arch/sparc64/solaris/fs.c
--- linux-2.6.8.1/arch/sparc64/solaris/fs.c	2004-08-14 12:55:09.000000000 +0200
+++ linux-2.6.8.1-bme0.05/arch/sparc64/solaris/fs.c	2004-08-18 11:36:34.000000000 +0200
@@ -362,7 +362,7 @@ static int report_statvfs(struct vfsmoun
 		int j = strlen (p);
 		
 		if (j > 15) j = 15;
-		if (IS_RDONLY(inode)) i = 1;
+		if (IS_RDONLY(inode, mnt)) i = 1;
 		if (mnt->mnt_flags & MNT_NOSUID) i |= 2;
 		if (!sysv_valid_dev(inode->i_sb->s_dev))
 			return -EOVERFLOW;
@@ -398,7 +398,7 @@ static int report_statvfs64(struct vfsmo
 		int j = strlen (p);
 		
 		if (j > 15) j = 15;
-		if (IS_RDONLY(inode)) i = 1;
+		if (IS_RDONLY(inode, mnt)) i = 1;
 		if (mnt->mnt_flags & MNT_NOSUID) i |= 2;
 		if (!sysv_valid_dev(inode->i_sb->s_dev))
 			return -EOVERFLOW;
diff -NurpP --minimal linux-2.6.8.1/fs/ext2/acl.c linux-2.6.8.1-bme0.05/fs/ext2/acl.c
--- linux-2.6.8.1/fs/ext2/acl.c	2004-08-14 12:55:59.000000000 +0200
+++ linux-2.6.8.1-bme0.05/fs/ext2/acl.c	2004-08-18 11:36:34.000000000 +0200
@@ -291,7 +291,7 @@ ext2_permission(struct inode *inode, int
 	int mode = inode->i_mode;
 
 	/* Nobody gets write access to a read-only fs */
-	if ((mask & MAY_WRITE) && IS_RDONLY(inode) &&
+	if ((mask & MAY_WRITE) && IS_RDONLY(inode, nd->mnt) &&
 	    (S_ISREG(mode) || S_ISDIR(mode) || S_ISLNK(mode)))
 		return -EROFS;
 	/* Nobody gets write access to an immutable file */
diff -NurpP --minimal linux-2.6.8.1/fs/ext2/ioctl.c linux-2.6.8.1-bme0.05/fs/ext2/ioctl.c
--- linux-2.6.8.1/fs/ext2/ioctl.c	2004-08-14 12:56:24.000000000 +0200
+++ linux-2.6.8.1-bme0.05/fs/ext2/ioctl.c	2004-08-18 11:36:34.000000000 +0200
@@ -29,7 +29,7 @@ int ext2_ioctl (struct inode * inode, st
 	case EXT2_IOC_SETFLAGS: {
 		unsigned int oldflags;
 
-		if (IS_RDONLY(inode))
+		if (IS_RDONLY(inode, filp->f_vfsmnt))
 			return -EROFS;
 
 		if ((current->fsuid != inode->i_uid) && !capable(CAP_FOWNER))
@@ -68,7 +68,7 @@ int ext2_ioctl (struct inode * inode, st
 	case EXT2_IOC_SETVERSION:
 		if ((current->fsuid != inode->i_uid) && !capable(CAP_FOWNER))
 			return -EPERM;
-		if (IS_RDONLY(inode))
+		if (IS_RDONLY(inode, filp->f_vfsmnt))
 			return -EROFS;
 		if (get_user(inode->i_generation, (int __user *) arg))
 			return -EFAULT;	
diff -NurpP --minimal linux-2.6.8.1/fs/ext2/xattr.c linux-2.6.8.1-bme0.05/fs/ext2/xattr.c
--- linux-2.6.8.1/fs/ext2/xattr.c	2004-08-14 12:55:35.000000000 +0200
+++ linux-2.6.8.1-bme0.05/fs/ext2/xattr.c	2004-08-18 11:36:34.000000000 +0200
@@ -496,7 +496,7 @@ ext2_xattr_set(struct inode *inode, int 
 	ea_idebug(inode, "name=%d.%s, value=%p, value_len=%ld",
 		  name_index, name, value, (long)value_len);
 
-	if (IS_RDONLY(inode))
+	if (IS_RDONLY_INODE(inode))
 		return -EROFS;
 	if (IS_IMMUTABLE(inode) || IS_APPEND(inode))
 		return -EPERM;
diff -NurpP --minimal linux-2.6.8.1/fs/ext3/acl.c linux-2.6.8.1-bme0.05/fs/ext3/acl.c
--- linux-2.6.8.1/fs/ext3/acl.c	2004-08-14 12:55:10.000000000 +0200
+++ linux-2.6.8.1-bme0.05/fs/ext3/acl.c	2004-08-18 11:36:34.000000000 +0200
@@ -296,7 +296,7 @@ ext3_permission(struct inode *inode, int
 	int mode = inode->i_mode;
 
 	/* Nobody gets write access to a read-only fs */
-	if ((mask & MAY_WRITE) && IS_RDONLY(inode) &&
+	if ((mask & MAY_WRITE) && IS_RDONLY(inode, nd->mnt) &&
 	    (S_ISREG(mode) || S_ISDIR(mode) || S_ISLNK(mode)))
 		return -EROFS;
 	/* Nobody gets write access to an immutable file */
diff -NurpP --minimal linux-2.6.8.1/fs/ext3/inode.c linux-2.6.8.1-bme0.05/fs/ext3/inode.c
--- linux-2.6.8.1/fs/ext3/inode.c	2004-08-14 12:55:47.000000000 +0200
+++ linux-2.6.8.1-bme0.05/fs/ext3/inode.c	2004-08-18 11:36:34.000000000 +0200
@@ -3056,7 +3056,7 @@ int ext3_change_inode_journal_flag(struc
 	 */
 
 	journal = EXT3_JOURNAL(inode);
-	if (is_journal_aborted(journal) || IS_RDONLY(inode))
+	if (is_journal_aborted(journal) || IS_RDONLY_INODE(inode))
 		return -EROFS;
 
 	journal_lock_updates(journal);
diff -NurpP --minimal linux-2.6.8.1/fs/ext3/ioctl.c linux-2.6.8.1-bme0.05/fs/ext3/ioctl.c
--- linux-2.6.8.1/fs/ext3/ioctl.c	2004-08-14 12:55:19.000000000 +0200
+++ linux-2.6.8.1-bme0.05/fs/ext3/ioctl.c	2004-08-18 11:36:34.000000000 +0200
@@ -34,7 +34,7 @@ int ext3_ioctl (struct inode * inode, st
 		unsigned int oldflags;
 		unsigned int jflag;
 
-		if (IS_RDONLY(inode))
+		if (IS_RDONLY(inode, filp->f_vfsmnt))
 			return -EROFS;
 
 		if ((current->fsuid != inode->i_uid) && !capable(CAP_FOWNER))
@@ -110,7 +110,7 @@ flags_err:
 
 		if ((current->fsuid != inode->i_uid) && !capable(CAP_FOWNER))
 			return -EPERM;
-		if (IS_RDONLY(inode))
+		if (IS_RDONLY(inode, filp->f_vfsmnt))
 			return -EROFS;
 		if (get_user(generation, (int __user *) arg))
 			return -EFAULT;
diff -NurpP --minimal linux-2.6.8.1/fs/ext3/xattr.c linux-2.6.8.1-bme0.05/fs/ext3/xattr.c
--- linux-2.6.8.1/fs/ext3/xattr.c	2004-08-14 12:55:33.000000000 +0200
+++ linux-2.6.8.1-bme0.05/fs/ext3/xattr.c	2004-08-18 11:36:34.000000000 +0200
@@ -496,7 +496,7 @@ ext3_xattr_set_handle(handle_t *handle, 
 	ea_idebug(inode, "name=%d.%s, value=%p, value_len=%ld",
 		  name_index, name, value, (long)value_len);
 
-	if (IS_RDONLY(inode))
+	if (IS_RDONLY_INODE(inode))
 		return -EROFS;
 	if (IS_IMMUTABLE(inode) || IS_APPEND(inode))
 		return -EPERM;
diff -NurpP --minimal linux-2.6.8.1/fs/fat/file.c linux-2.6.8.1-bme0.05/fs/fat/file.c
--- linux-2.6.8.1/fs/fat/file.c	2004-08-14 12:55:09.000000000 +0200
+++ linux-2.6.8.1-bme0.05/fs/fat/file.c	2004-08-18 11:36:34.000000000 +0200
@@ -91,7 +91,7 @@ void fat_truncate(struct inode *inode)
 	int nr_clusters;
 
 	/* Why no return value?  Surely the disk could fail... */
-	if (IS_RDONLY (inode))
+	if (IS_RDONLY_INODE(inode))
 		return /* -EPERM */;
 	if (IS_IMMUTABLE(inode))
 		return /* -EPERM */;
diff -NurpP --minimal linux-2.6.8.1/fs/hfsplus/ioctl.c linux-2.6.8.1-bme0.05/fs/hfsplus/ioctl.c
--- linux-2.6.8.1/fs/hfsplus/ioctl.c	2004-08-14 12:55:09.000000000 +0200
+++ linux-2.6.8.1-bme0.05/fs/hfsplus/ioctl.c	2004-08-18 11:36:34.000000000 +0200
@@ -33,7 +33,7 @@ int hfsplus_ioctl(struct inode *inode, s
 			flags |= EXT2_FLAG_NODUMP; /* EXT2_NODUMP_FL */
 		return put_user(flags, (int __user *)arg);
 	case HFSPLUS_IOC_EXT2_SETFLAGS: {
-		if (IS_RDONLY(inode))
+		if (IS_RDONLY(inode, filp->f_vfsmnt))
 			return -EROFS;
 
 		if ((current->fsuid != inode->i_uid) && !capable(CAP_FOWNER))
diff -NurpP --minimal linux-2.6.8.1/fs/inode.c linux-2.6.8.1-bme0.05/fs/inode.c
--- linux-2.6.8.1/fs/inode.c	2004-08-14 12:56:23.000000000 +0200
+++ linux-2.6.8.1-bme0.05/fs/inode.c	2004-08-18 11:36:34.000000000 +0200
@@ -1157,7 +1157,7 @@ void update_atime(struct inode *inode)
 		return;
 	if (IS_NODIRATIME(inode) && S_ISDIR(inode->i_mode))
 		return;
-	if (IS_RDONLY(inode))
+	if (IS_RDONLY_INODE(inode))
 		return;
 
 	now = current_kernel_time();
@@ -1181,14 +1181,14 @@ EXPORT_SYMBOL(update_atime);
  *	When ctime_too is specified update the ctime too.
  */
 
-void inode_update_time(struct inode *inode, int ctime_too)
+void inode_update_time(struct inode *inode, struct vfsmount *mnt, int ctime_too)
 {
 	struct timespec now;
 	int sync_it = 0;
 
 	if (IS_NOCMTIME(inode))
 		return;
-	if (IS_RDONLY(inode))
+	if (IS_RDONLY(inode, mnt))
 		return;
 
 	now = current_kernel_time();
diff -NurpP --minimal linux-2.6.8.1/fs/jfs/acl.c linux-2.6.8.1-bme0.05/fs/jfs/acl.c
--- linux-2.6.8.1/fs/jfs/acl.c	2004-08-14 12:54:48.000000000 +0200
+++ linux-2.6.8.1-bme0.05/fs/jfs/acl.c	2004-08-18 11:36:34.000000000 +0200
@@ -136,7 +136,7 @@ int jfs_permission(struct inode * inode,
 		/*
 		 * Nobody gets write access to a read-only fs.
 		 */
-		if (IS_RDONLY(inode) &&
+		if (IS_RDONLY(inode, nd->mnt) &&
 		    (S_ISREG(mode) || S_ISDIR(mode) || S_ISLNK(mode)))
 			return -EROFS;
 
diff -NurpP --minimal linux-2.6.8.1/fs/jfs/xattr.c linux-2.6.8.1-bme0.05/fs/jfs/xattr.c
--- linux-2.6.8.1/fs/jfs/xattr.c	2004-08-14 12:55:10.000000000 +0200
+++ linux-2.6.8.1-bme0.05/fs/jfs/xattr.c	2004-08-18 11:36:34.000000000 +0200
@@ -725,7 +725,7 @@ static int can_set_system_xattr(struct i
 static int can_set_xattr(struct inode *inode, const char *name,
 			 const void *value, size_t value_len)
 {
-	if (IS_RDONLY(inode))
+	if (IS_RDONLY_INODE(inode))
 		return -EROFS;
 
 	if (IS_IMMUTABLE(inode) || IS_APPEND(inode) || S_ISLNK(inode->i_mode))
diff -NurpP --minimal linux-2.6.8.1/fs/namei.c linux-2.6.8.1-bme0.05/fs/namei.c
--- linux-2.6.8.1/fs/namei.c	2004-08-14 12:55:10.000000000 +0200
+++ linux-2.6.8.1-bme0.05/fs/namei.c	2004-08-18 11:36:34.000000000 +0200
@@ -167,7 +167,7 @@ int vfs_permission(struct inode * inode,
 		/*
 		 * Nobody gets write access to a read-only fs.
 		 */
-		if (IS_RDONLY(inode) &&
+		if (IS_RDONLY_INODE(inode) &&
 		    (S_ISREG(mode) || S_ISDIR(mode) || S_ISLNK(mode)))
 			return -EROFS;
 
@@ -212,10 +212,14 @@ int permission(struct inode * inode,int 
 {
 	int retval;
 	int submask;
+	umode_t	mode = inode->i_mode;
 
 	/* Ordinary permission routines do not understand MAY_APPEND. */
 	submask = mask & ~MAY_APPEND;
 
+	if (nd && (mask & MAY_WRITE) && MNT_IS_RDONLY(nd->mnt) &&
+		(S_ISREG(mode) || S_ISDIR(mode) || S_ISLNK(mode)))
+		return -EROFS;
 	if (inode->i_op && inode->i_op->permission)
 		retval = inode->i_op->permission(inode, submask, nd);
 	else
@@ -1136,6 +1140,24 @@ static inline int may_create(struct inod
 	return permission(dir,MAY_WRITE | MAY_EXEC, nd);
 }
 
+static inline int mnt_may_create(struct vfsmount *mnt, struct inode *dir, struct dentry *child) {
+       if (child->d_inode)
+               return -EEXIST;
+       if (IS_DEADDIR(dir))
+               return -ENOENT;
+       if (mnt->mnt_flags & MNT_RDONLY)
+               return -EROFS;
+       return 0;
+}
+
+static inline int mnt_may_unlink(struct vfsmount *mnt, struct inode *dir, struct dentry *child) {
+       if (!child->d_inode)
+               return -ENOENT;
+       if (mnt->mnt_flags & MNT_RDONLY)
+               return -EROFS;
+       return 0;
+}
+
 /* 
  * Special case: O_CREAT|O_EXCL implies O_NOFOLLOW for security
  * reasons.
@@ -1257,7 +1279,7 @@ int may_open(struct nameidata *nd, int a
 			return -EACCES;
 
 		flag &= ~O_TRUNC;
-	} else if (IS_RDONLY(inode) && (flag & FMODE_WRITE))
+	} else if (IS_RDONLY(inode, nd->mnt) && (flag & FMODE_WRITE))
 		return -EROFS;
 	/*
 	 * An append-only file must be opened in append mode for writing.
@@ -1495,23 +1517,28 @@ do_link:
 struct dentry *lookup_create(struct nameidata *nd, int is_dir)
 {
 	struct dentry *dentry;
+	int error;
 
 	down(&nd->dentry->d_inode->i_sem);
-	dentry = ERR_PTR(-EEXIST);
+	error = -EEXIST;
 	if (nd->last_type != LAST_NORM)
-		goto fail;
+		goto out;
 	nd->flags &= ~LOOKUP_PARENT;
 	dentry = lookup_hash(&nd->last, nd->dentry);
 	if (IS_ERR(dentry))
+		goto ret;
+	error = mnt_may_create(nd->mnt, nd->dentry->d_inode, dentry);
+	if (error)
 		goto fail;
+	error = -ENOENT;
 	if (!is_dir && nd->last.name[nd->last.len] && !dentry->d_inode)
-		goto enoent;
+		goto fail;
+ret:
 	return dentry;
-enoent:
-	dput(dentry);
-	dentry = ERR_PTR(-ENOENT);
 fail:
-	return dentry;
+	dput(dentry);
+out:
+	return ERR_PTR(error);
 }
 
 int vfs_mknod(struct inode *dir, struct dentry *dentry, int mode, dev_t dev)
@@ -1740,7 +1767,11 @@ asmlinkage long sys_rmdir(const char __u
 	dentry = lookup_hash(&nd.last, nd.dentry);
 	error = PTR_ERR(dentry);
 	if (!IS_ERR(dentry)) {
+		error = mnt_may_unlink(nd.mnt, nd.dentry->d_inode, dentry);
+		if (error)
+			goto exit2;
 		error = vfs_rmdir(nd.dentry->d_inode, dentry);
+	exit2:
 		dput(dentry);
 	}
 	up(&nd.dentry->d_inode->i_sem);
@@ -1812,6 +1843,9 @@ asmlinkage long sys_unlink(const char __
 		/* Why not before? Because we want correct error value */
 		if (nd.last.name[nd.last.len])
 			goto slashes;
+		error = mnt_may_unlink(nd.mnt, nd.dentry->d_inode, dentry);
+		if (error)
+			goto exit2;
 		inode = dentry->d_inode;
 		if (inode)
 			atomic_inc(&inode->i_count);
@@ -2176,6 +2210,9 @@ static inline int do_rename(const char *
 	error = -EINVAL;
 	if (old_dentry == trap)
 		goto exit4;
+	error = -EROFS;
+	if (MNT_IS_RDONLY(newnd.mnt))
+		goto exit4;
 	new_dentry = lookup_hash(&newnd.last, new_dir);
 	error = PTR_ERR(new_dentry);
 	if (IS_ERR(new_dentry))
diff -NurpP --minimal linux-2.6.8.1/fs/namespace.c linux-2.6.8.1-bme0.05/fs/namespace.c
--- linux-2.6.8.1/fs/namespace.c	2004-08-14 12:55:35.000000000 +0200
+++ linux-2.6.8.1-bme0.05/fs/namespace.c	2004-08-18 11:36:34.000000000 +0200
@@ -221,37 +221,40 @@ static int show_vfsmnt(struct seq_file *
 	struct vfsmount *mnt = v;
 	int err = 0;
 	static struct proc_fs_info {
-		int flag;
-		char *str;
+		int s_flag;
+		int mnt_flag;
+		char *set_str;
+		char *unset_str;
 	} fs_info[] = {
-		{ MS_SYNCHRONOUS, ",sync" },
-		{ MS_DIRSYNC, ",dirsync" },
-		{ MS_MANDLOCK, ",mand" },
-		{ MS_NOATIME, ",noatime" },
-		{ MS_NODIRATIME, ",nodiratime" },
-		{ 0, NULL }
-	};
-	static struct proc_fs_info mnt_info[] = {
-		{ MNT_NOSUID, ",nosuid" },
-		{ MNT_NODEV, ",nodev" },
-		{ MNT_NOEXEC, ",noexec" },
-		{ 0, NULL }
+		{ MS_RDONLY, MNT_RDONLY, "ro", "rw" },
+		{ MS_SYNCHRONOUS, 0, ",sync", NULL },
+		{ MS_DIRSYNC, 0, ",dirsync", NULL },
+		{ MS_MANDLOCK, 0, ",mand", NULL },
+		{ MS_NOATIME, MNT_NOATIME, ",noatime", NULL },
+		{ MS_NODIRATIME, MNT_NODIRATIME, ",nodiratime", NULL },
+		{ 0, MNT_NOSUID, ",nosuid", NULL },
+		{ 0, MNT_NODEV, ",nodev", NULL },
+		{ 0, MNT_NOEXEC, ",noexec", NULL },
+		{ 0, 0, NULL, NULL }
 	};
-	struct proc_fs_info *fs_infop;
+	struct proc_fs_info *p;
+	unsigned long s_flags = mnt->mnt_sb->s_flags;
+	int mnt_flags = mnt->mnt_flags;
 
 	mangle(m, mnt->mnt_devname ? mnt->mnt_devname : "none");
 	seq_putc(m, ' ');
 	seq_path(m, mnt, mnt->mnt_root, " \t\n\\");
 	seq_putc(m, ' ');
 	mangle(m, mnt->mnt_sb->s_type->name);
-	seq_puts(m, mnt->mnt_sb->s_flags & MS_RDONLY ? " ro" : " rw");
-	for (fs_infop = fs_info; fs_infop->flag; fs_infop++) {
-		if (mnt->mnt_sb->s_flags & fs_infop->flag)
-			seq_puts(m, fs_infop->str);
-	}
-	for (fs_infop = mnt_info; fs_infop->flag; fs_infop++) {
-		if (mnt->mnt_flags & fs_infop->flag)
-			seq_puts(m, fs_infop->str);
+	seq_putc(m, ' ');
+	for (p = fs_info; (p->s_flag | p->mnt_flag) ; p++) {
+		if ((s_flags & p->s_flag) || (mnt_flags & p->mnt_flag)) {
+			if (p->set_str)
+				seq_puts(m, p->set_str);
+		} else {
+			if (p->unset_str)
+				seq_puts(m, p->unset_str);
+		}
 	}
 	if (mnt->mnt_sb->s_op->show_options)
 		err = mnt->mnt_sb->s_op->show_options(m, mnt);
@@ -618,11 +621,13 @@ out_unlock:
 /*
  * do loopback mount.
  */
-static int do_loopback(struct nameidata *nd, char *old_name, int recurse)
+static int do_loopback(struct nameidata *nd, char *old_name, unsigned long flags, int mnt_flags)
 {
 	struct nameidata old_nd;
 	struct vfsmount *mnt = NULL;
+	int recurse = flags & MS_REC;
 	int err = mount_is_safe(nd);
+
 	if (err)
 		return err;
 	if (!old_name || !*old_name)
@@ -654,6 +659,7 @@ static int do_loopback(struct nameidata 
 			spin_unlock(&vfsmount_lock);
 		} else
 			mntput(mnt);
+		mnt->mnt_flags = mnt_flags;
 	}
 
 	up_write(&current->namespace->sem);
@@ -999,12 +1005,18 @@ long do_mount(char * dev_name, char * di
 		((char *)data_page)[PAGE_SIZE - 1] = 0;
 
 	/* Separate the per-mountpoint flags */
+	if (flags & MS_RDONLY)
+		mnt_flags |= MNT_RDONLY;
 	if (flags & MS_NOSUID)
 		mnt_flags |= MNT_NOSUID;
 	if (flags & MS_NODEV)
 		mnt_flags |= MNT_NODEV;
 	if (flags & MS_NOEXEC)
 		mnt_flags |= MNT_NOEXEC;
+	if (flags & MS_NOATIME)
+		mnt_flags |= MNT_NOATIME;
+	if (flags & MS_NODIRATIME)
+		mnt_flags |= MNT_NODIRATIME;
 	flags &= ~(MS_NOSUID|MS_NOEXEC|MS_NODEV|MS_ACTIVE);
 
 	/* ... and get the mountpoint */
@@ -1020,7 +1032,7 @@ long do_mount(char * dev_name, char * di
 		retval = do_remount(&nd, flags & ~MS_REMOUNT, mnt_flags,
 				    data_page);
 	else if (flags & MS_BIND)
-		retval = do_loopback(&nd, dev_name, flags & MS_REC);
+		retval = do_loopback(&nd, dev_name, flags, mnt_flags);
 	else if (flags & MS_MOVE)
 		retval = do_move_mount(&nd, dev_name);
 	else
diff -NurpP --minimal linux-2.6.8.1/fs/ncpfs/file.c linux-2.6.8.1-bme0.05/fs/ncpfs/file.c
--- linux-2.6.8.1/fs/ncpfs/file.c	2004-08-14 12:54:48.000000000 +0200
+++ linux-2.6.8.1-bme0.05/fs/ncpfs/file.c	2004-08-18 11:36:34.000000000 +0200
@@ -175,7 +175,7 @@ ncp_file_read(struct file *file, char __
 
 	*ppos = pos;
 
-	if (!IS_RDONLY(inode)) {
+	if (!IS_RDONLY(inode, file->f_vfsmnt)) {
 		inode->i_atime = CURRENT_TIME;
 	}
 	
diff -NurpP --minimal linux-2.6.8.1/fs/ncpfs/mmap.c linux-2.6.8.1-bme0.05/fs/ncpfs/mmap.c
--- linux-2.6.8.1/fs/ncpfs/mmap.c	2004-08-14 12:55:47.000000000 +0200
+++ linux-2.6.8.1-bme0.05/fs/ncpfs/mmap.c	2004-08-18 11:36:34.000000000 +0200
@@ -123,7 +123,7 @@ int ncp_mmap(struct file *file, struct v
 	if (((vma->vm_end - vma->vm_start) >> PAGE_SHIFT) + vma->vm_pgoff 
 	   > (1U << (32 - PAGE_SHIFT)))
 		return -EFBIG;
-	if (!IS_RDONLY(inode)) {
+	if (!IS_RDONLY(inode, file->f_vfsmnt)) {
 		inode->i_atime = CURRENT_TIME;
 	}
 
diff -NurpP --minimal linux-2.6.8.1/fs/nfs/dir.c linux-2.6.8.1-bme0.05/fs/nfs/dir.c
--- linux-2.6.8.1/fs/nfs/dir.c	2004-08-14 12:55:32.000000000 +0200
+++ linux-2.6.8.1-bme0.05/fs/nfs/dir.c	2004-08-18 11:36:34.000000000 +0200
@@ -778,7 +778,8 @@ static int is_atomic_open(struct inode *
 	if (nd->flags & LOOKUP_DIRECTORY)
 		return 0;
 	/* Are we trying to write to a read only partition? */
-	if (IS_RDONLY(dir) && (nd->intent.open.flags & (O_CREAT|O_TRUNC|FMODE_WRITE)))
+	if (IS_RDONLY(dir, nd->mnt) &&
+		(nd->intent.open.flags & (O_CREAT|O_TRUNC|FMODE_WRITE)))
 		return 0;
 	return 1;
 }
@@ -1514,7 +1515,7 @@ nfs_permission(struct inode *inode, int 
 		 * Nobody gets write access to a read-only fs.
 		 *
 		 */
-		if (IS_RDONLY(inode) &&
+		if (IS_RDONLY(inode, nd->mnt) &&
 		    (S_ISREG(mode) || S_ISDIR(mode) || S_ISLNK(mode)))
 			return -EROFS;
 
diff -NurpP --minimal linux-2.6.8.1/fs/nfsd/vfs.c linux-2.6.8.1-bme0.05/fs/nfsd/vfs.c
--- linux-2.6.8.1/fs/nfsd/vfs.c	2004-08-14 12:55:19.000000000 +0200
+++ linux-2.6.8.1-bme0.05/fs/nfsd/vfs.c	2004-08-18 11:36:34.000000000 +0200
@@ -1556,7 +1556,7 @@ nfsd_permission(struct svc_export *exp, 
 	 */
 	if (!(acc & MAY_LOCAL_ACCESS))
 		if (acc & (MAY_WRITE | MAY_SATTR | MAY_TRUNC)) {
-			if (EX_RDONLY(exp) || IS_RDONLY(inode))
+			if (EX_RDONLY(exp) || IS_RDONLY(inode, exp->ex_mnt))
 				return nfserr_rofs;
 			if (/* (acc & MAY_WRITE) && */ IS_IMMUTABLE(inode))
 				return nfserr_perm;
diff -NurpP --minimal linux-2.6.8.1/fs/ntfs/inode.c linux-2.6.8.1-bme0.05/fs/ntfs/inode.c
--- linux-2.6.8.1/fs/ntfs/inode.c	2004-08-14 12:56:15.000000000 +0200
+++ linux-2.6.8.1-bme0.05/fs/ntfs/inode.c	2004-08-18 11:36:34.000000000 +0200
@@ -925,7 +925,7 @@ skip_large_dir_stuff:
 		/* Everyone gets read and scan permissions. */
 		vi->i_mode |= S_IRUGO | S_IXUGO;
 		/* If not read-only, set write permissions. */
-		if (!IS_RDONLY(vi))
+		if (!IS_RDONLY_INODE(vi))
 			vi->i_mode |= S_IWUGO;
 		/*
 		 * Apply the directory permissions mask set in the mount
@@ -1066,7 +1066,7 @@ no_data_attr_special_case:
 		/* Everyone gets all permissions. */
 		vi->i_mode |= S_IRWXUGO;
 		/* If read-only, noone gets write permissions. */
-		if (IS_RDONLY(vi))
+		if (IS_RDONLY_INODE(vi))
 			vi->i_mode &= ~S_IWUGO;
 		/* Apply the file permissions mask set in the mount options. */
 		vi->i_mode &= ~vol->fmask;
diff -NurpP --minimal linux-2.6.8.1/fs/open.c linux-2.6.8.1-bme0.05/fs/open.c
--- linux-2.6.8.1/fs/open.c	2004-08-14 12:54:48.000000000 +0200
+++ linux-2.6.8.1-bme0.05/fs/open.c	2004-08-18 11:36:34.000000000 +0200
@@ -239,7 +239,7 @@ static inline long do_sys_truncate(const
 		goto dput_and_out;
 
 	error = -EROFS;
-	if (IS_RDONLY(inode))
+	if (IS_RDONLY(inode, nd.mnt))
 		goto dput_and_out;
 
 	error = -EPERM;
@@ -363,7 +363,7 @@ asmlinkage long sys_utime(char __user * 
 	inode = nd.dentry->d_inode;
 
 	error = -EROFS;
-	if (IS_RDONLY(inode))
+	if (IS_RDONLY(inode, nd.mnt))
 		goto dput_and_out;
 
 	/* Don't worry, the checks are done in inode_change_ok() */
@@ -420,7 +420,7 @@ long do_utimes(char __user * filename, s
 	inode = nd.dentry->d_inode;
 
 	error = -EROFS;
-	if (IS_RDONLY(inode))
+	if (IS_RDONLY(inode, nd.mnt))
 		goto dput_and_out;
 
 	/* Don't worry, the checks are done in inode_change_ok() */
@@ -502,7 +502,8 @@ asmlinkage long sys_access(const char __
 	if (!res) {
 		res = permission(nd.dentry->d_inode, mode, &nd);
 		/* SuS v2 requires we report a read only fs too */
-		if(!res && (mode & S_IWOTH) && IS_RDONLY(nd.dentry->d_inode)
+		if(!res && (mode & S_IWOTH)
+		   && IS_RDONLY(nd.dentry->d_inode, nd.mnt)
 		   && !special_file(nd.dentry->d_inode->i_mode))
 			res = -EROFS;
 		path_release(&nd);
@@ -608,7 +609,7 @@ asmlinkage long sys_fchmod(unsigned int 
 	inode = dentry->d_inode;
 
 	err = -EROFS;
-	if (IS_RDONLY(inode))
+	if (IS_RDONLY(inode, file->f_vfsmnt))
 		goto out_putf;
 	err = -EPERM;
 	if (IS_IMMUTABLE(inode) || IS_APPEND(inode))
@@ -640,7 +641,7 @@ asmlinkage long sys_chmod(const char __u
 	inode = nd.dentry->d_inode;
 
 	error = -EROFS;
-	if (IS_RDONLY(inode))
+	if (IS_RDONLY(inode, nd.mnt))
 		goto dput_and_out;
 
 	error = -EPERM;
@@ -661,7 +662,8 @@ out:
 	return error;
 }
 
-static int chown_common(struct dentry * dentry, uid_t user, gid_t group)
+static int chown_common(struct dentry *dentry, struct vfsmount *mnt,
+	uid_t user, gid_t group)
 {
 	struct inode * inode;
 	int error;
@@ -673,7 +675,7 @@ static int chown_common(struct dentry * 
 		goto out;
 	}
 	error = -EROFS;
-	if (IS_RDONLY(inode))
+	if (IS_RDONLY(inode, mnt))
 		goto out;
 	error = -EPERM;
 	if (IS_IMMUTABLE(inode) || IS_APPEND(inode))
@@ -703,7 +705,7 @@ asmlinkage long sys_chown(const char __u
 
 	error = user_path_walk(filename, &nd);
 	if (!error) {
-		error = chown_common(nd.dentry, user, group);
+		error = chown_common(nd.dentry, nd.mnt, user, group);
 		path_release(&nd);
 	}
 	return error;
@@ -716,7 +718,7 @@ asmlinkage long sys_lchown(const char __
 
 	error = user_path_walk_link(filename, &nd);
 	if (!error) {
-		error = chown_common(nd.dentry, user, group);
+		error = chown_common(nd.dentry, nd.mnt, user, group);
 		path_release(&nd);
 	}
 	return error;
@@ -730,7 +732,7 @@ asmlinkage long sys_fchown(unsigned int 
 
 	file = fget(fd);
 	if (file) {
-		error = chown_common(file->f_dentry, user, group);
+		error = chown_common(file->f_dentry, file->f_vfsmnt, user, group);
 		fput(file);
 	}
 	return error;
diff -NurpP --minimal linux-2.6.8.1/fs/pipe.c linux-2.6.8.1-bme0.05/fs/pipe.c
--- linux-2.6.8.1/fs/pipe.c	2004-08-14 12:55:33.000000000 +0200
+++ linux-2.6.8.1-bme0.05/fs/pipe.c	2004-08-18 11:36:34.000000000 +0200
@@ -255,7 +255,7 @@ pipe_writev(struct file *filp, const str
 		kill_fasync(PIPE_FASYNC_READERS(*inode), SIGIO, POLL_IN);
 	}
 	if (ret > 0)
-		inode_update_time(inode, 1);	/* mtime and ctime */
+		inode_update_time(inode, filp->f_vfsmnt, 1);	/* mtime and ctime */
 	return ret;
 }
 
diff -NurpP --minimal linux-2.6.8.1/fs/reiserfs/file.c linux-2.6.8.1-bme0.05/fs/reiserfs/file.c
--- linux-2.6.8.1/fs/reiserfs/file.c	2004-08-14 12:54:47.000000000 +0200
+++ linux-2.6.8.1-bme0.05/fs/reiserfs/file.c	2004-08-18 11:36:34.000000000 +0200
@@ -1160,7 +1160,7 @@ ssize_t reiserfs_file_write( struct file
     if (res)
 	goto out;
 
-    inode_update_time(inode, 1); /* Both mtime and ctime */
+    inode_update_time(inode, file->f_vfsmnt, 1); /* Both mtime and ctime */
 
     // Ok, we are done with all the checks.
 
diff -NurpP --minimal linux-2.6.8.1/fs/reiserfs/ioctl.c linux-2.6.8.1-bme0.05/fs/reiserfs/ioctl.c
--- linux-2.6.8.1/fs/reiserfs/ioctl.c	2004-08-14 12:55:19.000000000 +0200
+++ linux-2.6.8.1-bme0.05/fs/reiserfs/ioctl.c	2004-08-18 11:36:34.000000000 +0200
@@ -38,7 +38,7 @@ int reiserfs_ioctl (struct inode * inode
 		i_attrs_to_sd_attrs( inode, ( __u16 * ) &flags );
 		return put_user(flags, (int __user *) arg);
 	case REISERFS_IOC_SETFLAGS: {
-		if (IS_RDONLY(inode))
+		if (IS_RDONLY(inode, filp->f_vfsmnt))
 			return -EROFS;
 
 		if ((current->fsuid != inode->i_uid) && !capable(CAP_FOWNER))
@@ -70,7 +70,7 @@ int reiserfs_ioctl (struct inode * inode
 	case REISERFS_IOC_SETVERSION:
 		if ((current->fsuid != inode->i_uid) && !capable(CAP_FOWNER))
 			return -EPERM;
-		if (IS_RDONLY(inode))
+		if (IS_RDONLY(inode, filp->f_vfsmnt))
 			return -EROFS;
 		if (get_user(inode->i_generation, (int __user *) arg))
 			return -EFAULT;	
diff -NurpP --minimal linux-2.6.8.1/fs/reiserfs/xattr.c linux-2.6.8.1-bme0.05/fs/reiserfs/xattr.c
--- linux-2.6.8.1/fs/reiserfs/xattr.c	2004-08-14 12:56:23.000000000 +0200
+++ linux-2.6.8.1-bme0.05/fs/reiserfs/xattr.c	2004-08-18 11:36:34.000000000 +0200
@@ -500,7 +500,7 @@ reiserfs_xattr_set (struct inode *inode,
     struct iattr newattrs;
     __u32 xahash = 0;
 
-    if (IS_RDONLY (inode))
+    if (IS_RDONLY_INODE (inode))
         return -EROFS;
 
     if (IS_IMMUTABLE (inode) || IS_APPEND (inode))
@@ -749,7 +749,7 @@ reiserfs_xattr_del (struct inode *inode,
     struct dentry *dir;
     int err;
 
-    if (IS_RDONLY (inode))
+    if (IS_RDONLY_INODE (inode))
         return -EROFS;
 
     dir = open_xa_dir (inode, FL_READONLY);
@@ -975,7 +975,7 @@ reiserfs_setxattr (struct dentry *dentry
         get_inode_sd_version (dentry->d_inode) == STAT_DATA_V1)
         return -EOPNOTSUPP;
 
-    if (IS_RDONLY (dentry->d_inode))
+    if (IS_RDONLY_INODE (dentry->d_inode))
         return -EROFS;
 
     if (IS_IMMUTABLE (dentry->d_inode) || IS_APPEND (dentry->d_inode))
@@ -1011,7 +1011,7 @@ reiserfs_removexattr (struct dentry *den
         get_inode_sd_version (dentry->d_inode) == STAT_DATA_V1)
         return -EOPNOTSUPP;
 
-    if (IS_RDONLY (dentry->d_inode))
+    if (IS_RDONLY_INODE (dentry->d_inode))
         return -EROFS;
 
     if (IS_IMMUTABLE (dentry->d_inode) || IS_APPEND (dentry->d_inode))
@@ -1342,7 +1342,7 @@ __reiserfs_permission (struct inode *ino
 		/*
 		 * Nobody gets write access to a read-only fs.
 		 */
-		if (IS_RDONLY(inode) &&
+		if (IS_RDONLY(inode, nd->mnt) &&
 		    (S_ISREG(mode) || S_ISDIR(mode) || S_ISLNK(mode)))
 			return -EROFS;
 
@@ -1397,7 +1397,9 @@ __reiserfs_permission (struct inode *ino
                 }
 #endif
 	} else {
+#ifdef CONFIG_REISERFS_FS_POSIX_ACL
 check_groups:
+#endif
 		if (in_group_p(inode->i_gid))
 			mode >>= 3;
 	}
@@ -1408,7 +1410,9 @@ check_groups:
 	if (((mode & mask & (MAY_READ|MAY_WRITE|MAY_EXEC)) == mask))
 		return 0;
 
+#ifdef CONFIG_REISERFS_FS_POSIX_ACL
 check_capabilities:
+#endif
 	/*
 	 * Read/write DACs are always overridable.
 	 * Executable DACs are overridable if at least one exec bit is set.
diff -NurpP --minimal linux-2.6.8.1/fs/xfs/xfs_inode.c linux-2.6.8.1-bme0.05/fs/xfs/xfs_inode.c
--- linux-2.6.8.1/fs/xfs/xfs_inode.c	2004-08-14 12:56:23.000000000 +0200
+++ linux-2.6.8.1-bme0.05/fs/xfs/xfs_inode.c	2004-08-18 11:36:34.000000000 +0200
@@ -3698,7 +3698,7 @@ xfs_iaccess(
 	if (mode & S_IWUSR) {
 		umode_t		imode = inode->i_mode;
 
-		if (IS_RDONLY(inode) &&
+		if (IS_RDONLY_INODE(inode) &&
 		    (S_ISREG(imode) || S_ISDIR(imode) || S_ISLNK(imode)))
 			return XFS_ERROR(EROFS);
 
diff -NurpP --minimal linux-2.6.8.1/include/linux/fs.h linux-2.6.8.1-bme0.05/include/linux/fs.h
--- linux-2.6.8.1/include/linux/fs.h	2004-08-14 12:55:09.000000000 +0200
+++ linux-2.6.8.1-bme0.05/include/linux/fs.h	2004-08-18 11:36:34.000000000 +0200
@@ -18,6 +18,7 @@
 #include <linux/cache.h>
 #include <linux/prio_tree.h>
 #include <linux/kobject.h>
+#include <linux/mount.h>
 #include <asm/atomic.h>
 
 struct iovec;
@@ -161,7 +162,8 @@ extern int leases_enable, dir_notify_ena
  */
 #define __IS_FLG(inode,flg) ((inode)->i_sb->s_flags & (flg))
 
-#define IS_RDONLY(inode) ((inode)->i_sb->s_flags & MS_RDONLY)
+#define IS_RDONLY_INODE(inode)	__IS_FLG(inode, MS_RDONLY)
+#define IS_RDONLY(inode, mnt)	(IS_RDONLY_INODE(inode) || MNT_IS_RDONLY(mnt))
 #define IS_SYNC(inode)		(__IS_FLG(inode, MS_SYNCHRONOUS) || \
 					((inode)->i_flags & S_SYNC))
 #define IS_DIRSYNC(inode)	(__IS_FLG(inode, MS_SYNCHRONOUS|MS_DIRSYNC) || \
@@ -981,8 +983,16 @@ static inline void mark_inode_dirty_sync
 
 static inline void touch_atime(struct vfsmount *mnt, struct dentry *dentry)
 {
-	/* per-mountpoint checks will go here */
-	update_atime(dentry->d_inode);
+	struct inode *inode = dentry->d_inode;
+
+	if (MNT_IS_NOATIME(mnt))
+		return;
+	if (S_ISDIR(inode->i_mode) && MNT_IS_NODIRATIME(mnt))
+		return;
+	if (IS_RDONLY(inode, mnt))
+		return;
+
+	update_atime(inode);
 }
 
 static inline void file_accessed(struct file *file)
@@ -1535,7 +1545,7 @@ extern ssize_t simple_read_from_buffer(v
 extern int inode_change_ok(struct inode *, struct iattr *);
 extern int __must_check inode_setattr(struct inode *, struct iattr *);
 
-extern void inode_update_time(struct inode *inode, int ctime_too);
+extern void inode_update_time(struct inode *inode, struct vfsmount *mnt, int ctime_too);
 
 static inline ino_t parent_ino(struct dentry *dentry)
 {
diff -NurpP --minimal linux-2.6.8.1/include/linux/mount.h linux-2.6.8.1-bme0.05/include/linux/mount.h
--- linux-2.6.8.1/include/linux/mount.h	2004-08-14 12:54:48.000000000 +0200
+++ linux-2.6.8.1-bme0.05/include/linux/mount.h	2004-08-18 11:36:34.000000000 +0200
@@ -17,6 +17,9 @@
 #define MNT_NOSUID	1
 #define MNT_NODEV	2
 #define MNT_NOEXEC	4
+#define MNT_RDONLY	8
+#define MNT_NOATIME	16
+#define MNT_NODIRATIME	32
 
 struct vfsmount
 {
@@ -36,6 +39,10 @@ struct vfsmount
 	struct namespace *mnt_namespace; /* containing namespace */
 };
 
+#define	MNT_IS_RDONLY(m)	((m)->mnt_flags & MNT_RDONLY)
+#define	MNT_IS_NOATIME(m)	((m)->mnt_flags & MNT_NOATIME)
+#define	MNT_IS_NODIRATIME(m)	((m)->mnt_flags & MNT_NODIRATIME)
+
 static inline struct vfsmount *mntget(struct vfsmount *mnt)
 {
 	if (mnt)
diff -NurpP --minimal linux-2.6.8.1/mm/filemap.c linux-2.6.8.1-bme0.05/mm/filemap.c
--- linux-2.6.8.1/mm/filemap.c	2004-08-14 12:56:25.000000000 +0200
+++ linux-2.6.8.1-bme0.05/mm/filemap.c	2004-08-18 11:36:34.000000000 +0200
@@ -1875,7 +1875,7 @@ generic_file_aio_write_nolock(struct kio
 	if (err)
 		goto out;
 
-	inode_update_time(inode, 1);
+	inode_update_time(inode, file->f_vfsmnt, 1);
 
 	/* coalesce the iovecs and go direct-to-BIO for O_DIRECT */
 	if (unlikely(file->f_flags & O_DIRECT)) {
